unit uEditDialog;
{$INCLUDE 'NamedFoldersLite.inc'}

interface

uses
  Windows, SysUtils, Classes,
  {$IFDEF FAR3} Plugin3, {$ELSE} PluginW, {$ENDIF} FarKeysW, FarColor,
  uLog, uFAR, uMessages, uConsts;

type
  TEditNamedFolderDialog = class(TFarDialog)
  private
    fName: string;
    fDirectory: string;
    fParam: string;
    fHostFile: string;
    fPluginID: TGUID;
    fDescription: string;
  private
    fEditName: integer;
    fEditDirectory: integer;
    fEditPluginID: integer;
    fEditParam: integer;
    fEditHostFile: integer;
    fEditDescription: integer;
    fButtonOK: integer;
    fButtonCancel: integer;
  private
    fEditNameBuf: TFarCharArray;
    fEditDirectoryBuf: TFarCharArray;
    fEditPluginIDBuf: TFarCharArray;
    fEditParamBuf: TFarCharArray;
    fEditHostFileBuf: TFarCharArray;
    fEditDescriptionBuf: TFarCharArray;
  protected
    function DialogProc(Msg, Param1, Param2: TIntPtr; var DlgProcResult: TIntPtr): boolean; override;
    function CanCloseDialog: boolean;
  public
    constructor Create;
    destructor Destroy; override;
    procedure ClearDialog; override;
    function Execute: boolean;
    property Name: string read fName write fName;
    property Directory: string read fDirectory write fDirectory;
    property PluginID: TGUID read fPluginID write fPluginID;
    property Param: string read fParam write fParam;
    property HostFile: string read fHostFile write fHostFile;
    property Description: string read fDescription write fDescription;
  end;

implementation

{ TEditNamedFolderDialog }

function TEditNamedFolderDialog.CanCloseDialog: boolean;
begin
  Result := True;
end;

procedure TEditNamedFolderDialog.ClearDialog;
begin
  inherited;
  fEditName := -1;
  fEditDirectory := -1;
  fEditPluginID := -1;
  fEditParam := -1;
  fEditHostFile := -1;
  fEditDescription := -1;
  fButtonOK := -1;
  fButtonCancel := -1;
end;

constructor TEditNamedFolderDialog.Create;
begin
  inherited Create( {$IFDEF FAR3_UP} PLUGIN_GUID, DIALOG_EDITNAMEDFOLDER_GUID {$ELSE} FarApi.ModuleNumber {$ENDIF} );
end;

destructor TEditNamedFolderDialog.Destroy;
begin
  inherited;
end;

function TEditNamedFolderDialog.DialogProc(Msg, Param1, Param2: TIntPtr; var DlgProcResult: TIntPtr): boolean;
var
  EditingData: PFarChar;
begin
  Result := False;
  if (Msg = DN_EDITCHANGE) and (Param1 = fEditName) and (fEditName >= 0) and (Param2 <> 0) then begin
    //Log('TEditNamedFolderDialog.DialogProc(DN_EDITCHANGE, fEditName, %d, %d)', [Param2, DlgProcResult]);
    EditingData := PFarDialogItem(Param2). {$IFDEF FAR3_UP} Data {$ELSE} PtrData {$ENDIF} ;
    if EditingData <> nil then begin
      //Log('- editing data: "%s"', [TFarUtils.CopyString(EditingData)]);
      DlgProcResult := 1;
      Result := True;
      while (DlgProcResult = 1) and (EditingData^ <> #0) do
        if (EditingData^ = '\') or (EditingData^ = '/') then
          DlgProcResult := 0
        else
          Inc(EditingData);
      if fButtonOK >= 0 then
        ItemEnabled[fButtonOK] := (DlgProcResult = 1) and CanCloseDialog;
    end;
    //Log('TEditNamedFolderDialog.DialogProc(DN_EDITCHANGE, fEditName, %d, %d) -> %d', [Param2, DlgProcResult, Integer(Result)]);
  end;
end;

function TEditNamedFolderDialog.Execute: boolean;
const
  GuidMask  = '{HHHHHHHH-HHHH-HHHH-HHHH-HHHHHHHHHHHH}';
  GuidEmpty = '{        -    -    -    -            }';
var
  Button: integer;
  PluginIDStr: string;
begin
  Result := False;
  // Build the dialog
                      AddDoubleBox(  3,  1, 62, 16, GetMsg(MTitle), 0);
                      AddLabel    (  5,  2,         GetMsg(MEditName), 0);
  fEditName :=        AddEdit     (  5,  3, 10,     Name, fEditNameBuf, nil, HISTORY_NAME, 32768, DIF_HISTORY {$IFDEF FAR3_UP} or DIF_FOCUS {$ENDIF} );
                      AddLabel    (  5,  4,         GetMsg(MEditDescription), 0);
  fEditDescription := AddEdit     (  5,  5, 54,     Description, fEditDescriptionBuf, nil, HISTORY_DESCRIPTION, 32768, DIF_HISTORY);
                      AddLabel    (  5,  6,         GetMsg(MEditDirectory), 0);
  fEditDirectory :=   AddEdit     (  5,  7, 54,     Directory, fEditDirectoryBuf, nil, HISTORY_DIRECTORY, 32768, DIF_HISTORY);
                      AddLabel    (  5,  8,         GetMsg(MEditPluginID), 0);
  fEditPluginID :=    AddEdit     (  5,  9, 40,     GUIDToString(PluginID), fEditPluginIDBuf, GuidMask, HISTORY_PLUGINID, 38, DIF_MASKEDIT or DIF_HISTORY);
                      AddLabel    (  5, 10,         GetMsg(MEditParam), 0);
  fEditParam :=       AddEdit     (  5, 11, 54,     Param, fEditParamBuf, nil, HISTORY_PARAM, 32768, DIF_HISTORY);
                      AddLabel    (  5, 12,         GetMsg(MEditHostFile), 0);
  fEditHostFile :=    AddEdit     (  5, 13, 54,     HostFile, fEditHostFileBuf, nil, HISTORY_HOSTFILE, 32768, DIF_HISTORY);
                      AddLabel    (  0, 14,         nil, DIF_SEPARATOR);
  fButtonOK :=        AddButton   (  0, 15,         GetMsg(MOk), DIF_CENTERGROUP);
  fButtonCancel :=    AddButton   (  0, 15,         GetMsg(MCancel), DIF_CENTERGROUP);
  // Set focus
  {$IFNDEF FAR3_UP}
  Items[fEditName].Focus := 1;
  {$ENDIF}
  // Set the default button
  {$IFDEF FAR3_UP}
  Items[fButtonOK].Flags := Items[fButtonOK].Flags or DIF_DEFAULTBUTTON;
  {$ELSE}
  Items[fButtonOK].DefaultButton := 1;
  {$ENDIF}
  // Show the dialog
  if Build(-1, -1, 66, 18, 0, HELP_EDITOR) then begin
    if inherited Execute(Button) then begin
      if Button = fButtonOK then begin
        Name := ItemText[fEditName];
        Directory := ItemText[fEditDirectory];
        PluginIDStr := ItemText[fEditPluginID];
        Param := ItemText[fEditParam];
        HostFile := ItemText[fEditHostFile];
        Description := ItemText[fEditDescription];
        if (PluginIDStr = '') or (PluginIDStr = GuidEmpty) then
          PluginID := GUID_NULL
        else
          PluginID := StringToGUID(PluginIDStr);
        Result := True;
      end;
    end;
  end;
end;

end.
