unit uFunctions;

interface

function ExtractPathElement(var Path: string; out Element: string): boolean;
function ExtractLastPathElement(var Path: string; out Element: string): boolean;
function GetLastPathElement(const Path: string): string;
function GetPath(const Path: string): string;
function GetAbsoluteDirectory(const CurrentDir, RelativeDir: string): string;

implementation

function GetLastPathElement(const Path: string): string;
var
  i: Integer;
begin
  if Path = '' then
    Result := ''
  else begin
    Result := Path;
    for i := Length(Path) downto 1 do
      if (Path[i] = '\') or (Path[i] = '/') then begin
        Result := Copy(Path, Succ(i), MaxInt);
        Break;
      end;
  end;
end;

function ExtractLastPathElement(var Path: string; out Element: string): boolean;
var
  i: Integer;
begin
  if Path = '' then
    Result := False
  else begin
    Result := True;
    for i := Length(Path) downto 1 do
      if (Path[i] = '\') or (Path[i] = '/') then begin
        Element := Copy(Path, Succ(i), MaxInt);
        SetLength(Path, Pred(i));
        Exit;
      end;
    Element := Path;
    Path := '';
  end;
end;

function GetPath(const Path: string): string;
var
  i: integer;
begin
  Result := '';
  for i := Length(Path) downto 1 do begin
    if (Path[i] = '\') or (Path[i] = '/') then begin
      Result := Copy(Path, 1, i);
      Break;
    end;
  end;
end;

function ExtractPathElement(var Path: string; out Element: string): boolean;
var
  i: integer;
begin
  if Path = '' then
    Result := False
  else begin
    Result := True;
    for i := 1 to Length(Path) do
      if (Path[i] = '\') or (Path[i] = '/') then begin
        Element := Copy(Path, 1, Pred(i));
        Delete(Path, 1, i);
        Exit;
      end;
    Element := Path;
    Path := '';
  end;
end;

function GetAbsoluteDirectory(const CurrentDir, RelativeDir: string): string;
var
  RelDir, Dir, Dummy: string;
begin
  Result := CurrentDir;
  if RelativeDir <> '' then begin
    if (RelativeDir[1] = '\') or (RelativeDir[1] = '/') then
      Result := '';
    RelDir := RelativeDir;
    while ExtractPathElement(RelDir, Dir) do
      if Dir <> '' then
        if Dir = '.' then
          // Do nothing
        else if Dir = '..' then
          ExtractLastPathElement(Result, Dummy)
        else
          if Result = '' then
            Result := Dir
          else
            Result := Result + '/' + Dir;
  end;
end;

end.
