unit uOptionsDialog;
{$INCLUDE 'NamedFoldersLite.inc'}

interface

uses
  SysUtils, Windows, Classes,
  {$IFDEF FAR3} Plugin3, {$ELSE} PluginW, {$ENDIF} FarKeysW, FarColor,
  uSystem, uFAR, uLog,
  uConsts, uMessages, uOptions;

type
  TOptionsDialog = class(TFarDialog)
  private
    fOptions: TOptions;
  private
    fCheckSaveLastPanelDirectory: integer;
    fCheckUseSubstEngines: integer;
    fCheckUseSubstEngine_KnownFolders: integer;
    fCheckUseSubstEngine_Environment: integer;
    fCheckUseSubstEngine_BatchEnvironment: integer;
    fCheckUseSubstEngine_Registry: integer;
    fCheckAllowPathSuffix: integer;
    fCheckAllowPluginCalls: integer;
    fCheckAllowFuzzySearch: integer;
    fCheckAllowSearchOverDirectories: integer;
    fButtonOK: integer;
    fButtonCancel: integer;
  protected
  public
    class function ExecuteDialog: boolean;
    constructor Create;
    destructor Destroy; override;
    procedure ClearDialog; override;
    function Execute: boolean;
    property Options: TOptions read fOptions write fOptions;
  end;

implementation

{ TOptionsDialog }

procedure TOptionsDialog.ClearDialog;
begin
  inherited;
  fCheckSaveLastPanelDirectory := -1;
  fCheckUseSubstEngines := -1;
  fCheckUseSubstEngine_KnownFolders := -1;
  fCheckUseSubstEngine_Environment := -1;
  fCheckUseSubstEngine_BatchEnvironment := -1;
  fCheckUseSubstEngine_Registry := -1;
  fCheckAllowPathSuffix := -1;
  fCheckAllowPluginCalls := -1;
  fCheckAllowFuzzySearch := -1;
  fCheckAllowFuzzySearch := -1;
  fButtonOK := -1;
  fButtonCancel := -1;
end;

constructor TOptionsDialog.Create;
begin
  inherited Create( {$IFDEF FAR3_UP} PLUGIN_GUID, DIALOG_CONFIG_GUID {$ELSE} FarApi.ModuleNumber {$ENDIF} );
end;

destructor TOptionsDialog.Destroy;
begin
  inherited;
end;

class function TOptionsDialog.ExecuteDialog: boolean;
var
  Options: TOptions;
  Dlg: TOptionsDialog;
begin
  Result := False;
  Options := TOptions.Create;
  try
    Dlg := TOptionsDialog.Create;
    try
      Dlg.Options := Options;
      if Dlg.Execute then
        Result := True;
    finally
      FreeAndNil(Dlg);
    end;
  finally
    FreeAndNil(Options);
  end;
end;

function TOptionsDialog.Execute: boolean;
const
  OPTION_LINES_COUNT = 10;
var
  Button: integer;
begin
  Result := False;
  // Build the dialog
                                            AddDoubleBox(  3,  1, 62,  OPTION_LINES_COUNT + 4, GetMsg(MTitle), 0);
  fCheckSaveLastPanelDirectory :=           AddCheckbox (  5,  2,                      GetMsg(MSaveLastPanelDirectory), Options.SaveLastPanelDirectory, 0);
  fCheckUseSubstEngines :=                  AddCheckbox (  5,  3,                      GetMsg(MUseSubstEngines), Options.UseSubstitutionEngines, 0);
  fCheckUseSubstEngine_KnownFolders :=      AddCheckbox (  7,  4,                      GetMsg(MUseSubstEngine_KnownFolders), Options.UseSubstitutionEngine_KnownFolders, 0);
  fCheckUseSubstEngine_Environment :=       AddCheckbox (  7,  5,                      GetMsg(MUseSubstEngine_Environment), Options.UseSubstitutionEngine_Environment, 0);
  fCheckUseSubstEngine_BatchEnvironment :=  AddCheckbox (  7,  6,                      GetMsg(MUseSubstEngine_BatchEnvironment), Options.UseSubstitutionEngine_BatchEnvironment, 0);
  fCheckUseSubstEngine_Registry :=          AddCheckbox (  7,  7,                      GetMsg(MUseSubstEngine_Registry), Options.UseSubstitutionEngine_Registry, 0);
  fCheckAllowPathSuffix :=                  AddCheckbox (  5,  8,                      GetMsg(MAllowPathSuffix), Options.AllowPathSuffix, 0);
  fCheckAllowPluginCalls :=                 AddCheckbox (  5,  9,                      GetMsg(MAllowPluginCalls), Options.AllowPluginCalls, 0);
  fCheckAllowFuzzySearch :=                 AddCheckbox (  5, 10,                      GetMsg(MAllowFuzzySearch), Options.AllowFuzzySearch, 0);
  fCheckAllowSearchOverDirectories :=       AddCheckbox (  5, 11,                      GetMsg(MAllowSearchOverDirectories), Options.AllowSearchOverDirectories, 0);
                                            AddLabel    (  0,  2 + OPTION_LINES_COUNT, nil, DIF_SEPARATOR);
  fButtonOK :=                              AddButton   (  0,  3 + OPTION_LINES_COUNT, GetMsg(MOk), DIF_CENTERGROUP);
  fButtonCancel :=                          AddButton   (  0,  3 + OPTION_LINES_COUNT, GetMsg(MCancel), DIF_CENTERGROUP);
  // Set the default button
  {$IFDEF FAR3_UP}
  Items[fButtonOK].Flags := Items[fButtonOK].Flags or DIF_DEFAULTBUTTON;
  {$ELSE}
  Items[fButtonOK].DefaultButton := 1;
  {$ENDIF}
  // Show the dialog
  if Build(-1, -1, 66, 6 + OPTION_LINES_COUNT, 0, 'Options') then begin
    if inherited Execute(Button) then begin
      if Button = fButtonOK then begin
        //Log('OK pressed, saving the settings');
        Options.SaveLastPanelDirectory := ItemChecked[fCheckSaveLastPanelDirectory];
        Options.UseSubstitutionEngines := ItemChecked[fCheckUseSubstEngines];
        Options.UseSubstitutionEngine_KnownFolders := ItemChecked[fCheckUseSubstEngine_KnownFolders];
        Options.UseSubstitutionEngine_Environment := ItemChecked[fCheckUseSubstEngine_Environment];
        Options.UseSubstitutionEngine_BatchEnvironment := ItemChecked[fCheckUseSubstEngine_BatchEnvironment];
        Options.UseSubstitutionEngine_Registry := ItemChecked[fCheckUseSubstEngine_Registry];
        Options.AllowPathSuffix := ItemChecked[fCheckAllowPathSuffix];
        Options.AllowPluginCalls := ItemChecked[fCheckAllowPluginCalls];
        Options.AllowFuzzySearch := ItemChecked[fCheckAllowFuzzySearch];
        Options.AllowSearchOverDirectories := ItemChecked[fCheckAllowSearchOverDirectories];
        Options.Save;
        Result := True;
      end;
    end;
  end;
end;

end.
